/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AddressbooksSynchronizer.
 *
 * The Initial Developer of the Original Code is Gnter Gersdorf.
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  Gnter Gersdorf <G.Gersdorf@ggbs.de>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

var gTempName;

var pref = Components.classes["@mozilla.org/preferences-service;1"].
    getService(Components.interfaces.nsIPrefBranch);
var rdf = Components.classes["@mozilla.org/rdf/rdf-service;1"].
    getService(Components.interfaces.nsIRDFService);
var directory = rdf.GetResource("moz-abdirectory://").
    QueryInterface(Components.interfaces.nsIAbDirectory);
var gDirectoryService = Components.classes['@mozilla.org/file/directory_service;1'].
    getService(Components.interfaces.nsIProperties);


/*
 *  Read a mab file and return the data
 */
function readMabData(nsIMabFile, addLastMod)
{
  if (!nsIMabFile.exists()) return '';

  var fs = Components.classes['@mozilla.org/network/file-input-stream;1']
          .createInstance(Components.interfaces.nsIFileInputStream);
  var ss = Components.classes['@mozilla.org/scriptableinputstream;1']
          .createInstance(Components.interfaces.nsIScriptableInputStream);
  var str='';
  try{
    fs.init(nsIMabFile, 0x01, 0000, false);
    ss.init(fs);
    str = ss.read(-1);
    ss.close();
    fs.close();
  } catch(e){ alert('Read file:\n'+e); }
//    return str;
  if (addLastMod) {
    var lmt=new Date(nsIMabFile.lastModifiedTime);
    return str.replace(/-->\s?[^\r\n]*/,'--> '+lmt);//append to the // <!-- <mdb:mork:z v="1.4"/> -->
  } else
    return str;
}

/*
 *  Save Data to a mab File
 */
function writeMabData(nsIMabFile, abData)
{
  try {
    //if (nsIMabFile.exists() && nsIMabFile.isFile()) nsIMabFile.remove(false);
    var strm = Components.classes["@mozilla.org/network/file-output-stream;1"].
      createInstance(Components.interfaces.nsIFileOutputStream);
    strm.QueryInterface(Components.interfaces.nsIOutputStream);
    strm.QueryInterface(Components.interfaces.nsISeekableStream);
    strm.init( nsIMabFile, 0x20|0x02|0x08, 0600, 0 );
      // 0x02=PR_WRONLY, 0x08=PR_CREATE_FILE, 0x20=PR_TRUNCATE
    strm.write( abData, abData.length );
    strm.flush();
    strm.close();
  } catch(e) { alert('File write:\n'+e); }
  return nsIMabFile;
}

/*
 *  copy a mab file to another location
 */
function saveMabFile(mabName, mabFile, directory, filename)
{
    var localMabFile = getDir('ProfD');
    localMabFile.appendRelativePath( mabFile );

    try {
      var file = Components.classes["@mozilla.org/file/local;1"].
        createInstance(Components.interfaces.nsILocalFile);
      file.initWithFile( directory );
      file.appendRelativePath( filename );
//TODO: eventuell umbenennen statt lschen
      if (file.exists() && file.isFile()) { file.remove(false); }
      localMabFile.copyTo( directory, filename );
        //creates directory, if not existing! retains lastModifiedTime !
    } catch(e) { return false; }
    return true;
} //

/*
 *  creates/replaces address book from mab file
 */
function loadMabFile(mabName, mabFile, directory, filename, oldfilename, force, checkLastMod)
{
    var syncload=false;
    var file;
    if (directory) { // sync load
      syncload=true;
      if (filename) { // 'local' or 'remote'
        file = Components.classes["@mozilla.org/file/local;1"].
          createInstance(Components.interfaces.nsILocalFile);
        file.initWithFile( directory );
        file.appendRelativePath( filename );
      } else {  // 'imap'
        file=directory;
      }
    } else {  // single load
      file = Components.classes["@mozilla.org/file/local;1"].
        createInstance(Components.interfaces.nsILocalFile);
      try{ file.initWithPath( filename ); }
      catch(e) {
        alert(strings.getString('nofile'));
        return false;
      }
    }
    if (!file.exists()) {
      if (syncload && oldfilename) { // file possibly stored with old version
        filename=oldfilename
        file.initWithFile( directory );
        file.appendRelativePath( oldfilename );
      }
      if (!file.exists()) {
        if (syncload) { // file possible store with very old version
          var veryoldfilename = mabName.replace( /\W/g, "_" )+".mab";
          if (veryoldfilename!=filename) {
            var file = Components.classes["@mozilla.org/file/local;1"].
              createInstance(Components.interfaces.nsILocalFile);
            file.initWithFile( directory );
            file.appendRelativePath( veryoldfilename );
          }
        }
      }
      if (!file.exists()) {
        alert(strings.getString('nofile'));
        return false;
      }
    }
    if (!force && checkLastMod) {
      var localMabFile = getDir('ProfD');
      localMabFile.appendRelativePath( mabFile );
      if (localMabFile.lastModifiedTime >= file.lastModifiedTime)
        return 'skipped';
    }

    var ret;
    if ((ret=replaceAddressBook(mabName, mabFile, file))) {
      alert(ret);
      return 'notcopied';
    } else {
      if (!syncload) deleteTempFile(gTempName);
      return 'copied';
    }
}


/*
 * returns list of mab filenames to be synced
 */
function syncedAddressbooks(direction)
{   //direction: up, down, all
  var preftail='';
  if (direction!='all') {
    var separateUpDown;
    try{ separateUpDown = pref.getBoolPref( "extensions.addressbookssync.separateupdown" ); }
      catch(e){ separateUpDown=false; }
    preftail=(separateUpDown?'.'+direction:'');
  }

  var mabs=new Object();
  var cn = directory.childNodes;
  var aBook = cn.getNext();
  while( aBook != null )
  {
    var cur = aBook.QueryInterface(Components.interfaces.nsIAbDirectory);
    var dirType = cur.directoryProperties.dirType;
    if (dirType==2) {            // ignore ldap-addressbooks (dirType!=2)
      var sync=false;
      if (direction!='all') {
        var prefName = cur.directoryProperties.prefName;
        try { sync = pref.getBoolPref( prefName+".extensions.addressbookssync"+preftail ); }
          catch(e) { sync=false; }
      } else {
        sync=true;
      }
      if (sync) {
        var file = cur.directoryProperties.fileName;
        if (mabs[file])
          dump('nsIAbDirectory::childNodes returned double addressbook '+cur.directoryProperties.description+'('+file+')\n');
        else
          mabs[file]=cur.directoryProperties;
      }
    }
    try { aBook = cn.getNext(); } catch(e) { break; }
  } // while
  if (direction=='all') return mabs;
  var mabFiles=new Array();
  for (mab in mabs) mabFiles.push(mab);
//TODO: probably should return the mabs-Object
  return mabFiles;
}

/*
 * return address book data structure associated with filename or description (=address book name)
 */
function getAddressBook( name )
{
    var cn = directory.childNodes;
    while( cn.hasMoreElements() )
    {
      var cur = cn.getNext().QueryInterface(Components.interfaces.nsIAbDirectory);
      if( cur.directoryProperties.fileName == name || cur.directoryProperties.description == name)
          return cur;
    } // while
    return null;
} // getAddressBook

/*
 * create new address book
 */
function createAddressBook( name, fileName, syncstate )
{
  try {
    var abProps = Components.classes["@mozilla.org/addressbook/properties;1"].
      createInstance(Components.interfaces.nsIAbDirectoryProperties);
    abProps.description = name;
    abProps.dirType = 2;
    abProps.fileName = fileName;
    directory.createNewDirectory( abProps );
    var cur=getAddressBook(fileName);

    var p=syncstate[0];
    if (p=='b')
      pref.setBoolPref( cur.directoryProperties.prefName+".extensions.addressbookssync", true );
    else if (p=='_')
      pref.setBoolPref( cur.directoryProperties.prefName+".extensions.addressbookssync", false );
    p=syncstate[1];
    if (p=='u')
      pref.setBoolPref( cur.directoryProperties.prefName+".extensions.addressbookssync.up", true );
    else if (p=='_')
      pref.setBoolPref( cur.directoryProperties.prefName+".extensions.addressbookssync.up", false );
    p=syncstate[2];
    if (p=='d')
      pref.setBoolPref( cur.directoryProperties.prefName+".extensions.addressbookssync.down", true );
    else if (p=='_')
      pref.setBoolPref( cur.directoryProperties.prefName+".extensions.addressbookssync.down", false );

  } catch (e) { /*alert('createAddressBook('+name+','+fileName+')\n'+e);*/ }
} // createAddressBook

/*
 * delete an address book
 */
function deleteAddressBook( name, fileName )
{
  // Move focus away from the to be deleted addressbook
    // this has to be done in the AddressBook window :-((
//  try {  SetAbView(kPersonalAddressbookURI, false);} catch(e) { alert(e);}

  try {
    var syncstate;
    var cur=getAddressBook(fileName);
    if (cur && cur.directoryProperties.URI!=kPersonalAddressbookURI &&
          cur.directoryProperties.URI!=kCollectedAddressbookURI) {
        var prefName = cur.directoryProperties.prefName;
          //prefName='ldap_2.servers.name' mit name='history', 'pab' oder eigener
        var p;
        try { p = pref.getBoolPref( prefName+".extensions.addressbookssync" ); syncstate=(p?'b':'_'); }
          catch(e) { syncstate='.'; }
        try { p = pref.getBoolPref( prefName+".extensions.addressbookssync.up" ); syncstate+=(p?'u':'_'); }
          catch(e) { syncstate+='.'; }
        try { p = pref.getBoolPref( prefName+".extensions.addressbookssync.down" ); syncstate+=(p?'d':'_'); }
          catch(e) { syncstate+='.'; }
        directory.deleteDirectory( cur );
        pref.deleteBranch( prefName );
        pref.deleteBranch( prefName+'.position' );  // sometimes remains!?
        return syncstate;
    }
  } catch (e) { alert('deleteAddressBook('+name+','+fileName+')\n'+e); return false; }
} // deleteAddressBook


function replaceAddressBook(mabName, mabFile, file) {
if (!mabFile) {
 var err='Internal error: empty filename for addressbook "'+mabName+'"';
 alert(err);
 return err;
}
  var cur=getAddressBook(mabName);
  var ret;
  var syncpolicy;
  try{ syncpolicy = pref.getCharPref( "extensions.addressbookssync.syncpolicy" ); }
    catch(e){ syncpolicy='entry'; }
  if (syncpolicy=='entry') {
    //workaround for a very mysterious problem:
    // Maillists are not loaded if filename contains spaces!
    if (file.path.indexOf(' ')>=0) {
      var newFile=getDir('TmpD');
      gTempName='abstemp.mab';
      newFile.appendRelativePath(gTempName);
      if (newFile.exists() && newFile.isFile()) { newFile.remove(false); }
      file.copyTo( getDir('TmpD'), gTempName );
      file=newFile;
    }
    ret=replaceAddressBook_viaEntry(mabName, mabFile, file);
  } else {
    var uri=cur.directoryProperties.URI;
    if (uri==kPersonalAddressbookURI || uri==kCollectedAddressbookURI) {
      ret=replaceAddressBook_viaEntry(mabName, mabFile, file);
    } else {
      ret=replaceAddressBook_viaFile(mabName, mabFile, file);
    }
  }
  cur.lastModifiedDate=gAddressbooksSync.appstarttime; //set to 'not modified'
  return ret;
}

function replaceAddressBook_viaFile(mabName, mabFile, file)
{
  var syncstate=deleteAddressBook( mabName, mabFile ); // NOP for 'history' und 'pab'!
  statustxt('.',false, false);
  var nsIMabFile = getDir('ProfD');
  var path=nsIMabFile.path;
  nsIMabFile.appendRelativePath( mabFile );
  if (nsIMabFile.path==path) {
   var err='Internal error: error creating path to addressbook "'+mabName+'" with filename "'+mabFile+'"';
   alert(err);
   return err;
  }
  if (nsIMabFile.exists() && nsIMabFile.isFile()) { nsIMabFile.remove(false); }
  file.copyTo( getDir('ProfD'), mabFile );
  statustxt('.',false, false);

  createAddressBook( mabName, mabFile, syncstate );
  return '';
}

/*
 * replace content of an address book with given data
 */
function replaceAddressBook_viaEntry(mabName, mabFile, file)
{
  var cur=getAddressBook(mabFile);  //cur=nsIAbDirectory
  if (!cur) return 'failed to create addressbook';

  //
  // clear target addressbook
  //
  try  {
    var cardsE=cur.childCards;
    var cards = Components.classes["@mozilla.org/supports-array;1"].
      getService(Components.interfaces.nsISupportsArray);
      //convert nsIEnumerator to nsISupportsArray (don't try cur.dropCard()!)
    var goon=1;
      // Loop over the cards
    try { cardsE.first() } catch (e) { goon=0; }
    while (goon) {
      var card = cardsE.currentItem().QueryInterface(Components.interfaces.nsIAbCard);
      cards.AppendElement(card);
      try { cardsE.next() } catch(e) { goon=0; }
    }
    cur.deleteCards(cards);  // delete the cards from the to-be synced addressbook

    var lists=cur.childNodes;  //delete lists
    while( lists.hasMoreElements() ) {
      var list = lists.getNext().QueryInterface(Components.interfaces.nsIAbDirectory);
      cur.deleteDirectory(list);
    }
  } catch(e) { alert('clear addressbook: '+e); }

// The cards are not deleted immediately, they are put on a deletedCards Table
// This table can't be purge programmatically, they are automatically purged
// iff more than 50 Cards are in the table and only cards are purged that are
// older than 6 month :-(((
// Therefore I set the LastModifiedDate attribute on the deleted cards to an
// appropriate date (1.1.2000) so they will be deleted later on
  var ab = Components.classes["@mozilla.org/addressbook;1"].
    getService(Components.interfaces.nsIAddressBook);
  var curDB=ab.getAbDatabaseFromURI(cur.directoryProperties.URI);

  try  {
    var count=new Object();
    var dListObj=new Object();
    curDB.getDeletedCardList(count, dListObj);
    if (count.value) {
      var dList=dListObj.value;
      for (var i=0; i<dList.Count(); i++)
      {
        var card = dList.GetElementAt(i).QueryInterface(Components.interfaces.nsIAbCard);
//        386d3570 = 946681200 = 1.Jan.2000, must not be 0!!
        curDB.setCardValue(card, 'LastModifiedDate', '386d3570' , false);
//wow, with this function, i can set my own values!!
      }
    }
  } catch(e) { alert('reset modDate: '+e); }
  curDB.close(true);

  // open the addressbook database file
  var db=Components.classes["@mozilla.org/addressbook/carddatabase;1"].
    getService(Components.interfaces.nsIAddrDatabase);
  var tdb=db.open(file, true, false);
  tdb.QueryInterface(Components.interfaces.nsIAddrDatabase);
  //tdb.dbPath is undefined
  // give me the maillists
    // must be moz-abmdbdirectory, not moz-abdirectory
  var temp=Components.classes['@mozilla.org/rdf/resource-factory;1?name=moz-abmdbdirectory']
      .createInstance(Components.interfaces.nsIAbDirectory);
  tdb.getMailingListsFromDB(temp);

  var tCards=tdb.enumerateCards(temp);

  // create maillists
  try  {
    var tMailLists=temp.childNodes;
    while (tMailLists.hasMoreElements()) {
      var tMailList=tMailLists.getNext().QueryInterface(Components.interfaces.nsIAbDirectory);
      var nMailList = Components.classes["@mozilla.org/addressbook/directoryproperty;1"].
        createInstance().QueryInterface(Components.interfaces.nsIAbDirectory);
      nMailList.dirName = tMailList.dirName;
      nMailList.listNickName = tMailList.listNickName;
      nMailList.description = tMailList.description;
      cur.addMailList(nMailList);
    }
  }catch(e) { alert('create maillists: '+e); }

  //
  // copy cards from database to mab
  //
// - Thunderbird has no way to copy a maillist! If you drag and drop a maillist
//   inside the addressbook manager the maillist becomes a normal card!!
// - The dirName Property of maillists is unique! Thunderbird refuses to create
//   a maillist with the same name (applies only in the GUI)
// - => How to:
//   create maillists
//   loop over old cards
//     copy card
//     find all maillists, the card is in
//       find the corresponding new maillist
//       insert the copy of the card into these maillists

  try {
    var goon=1;
      // Loop over the cards
    try { tCards.first() } catch (e) { goon=0; }
    while (goon) {
      // need two references to the current card. One for a later use with .equals
      // the other gets modified to be the copied card
      var tCard = tCards.currentItem().QueryInterface(Components.interfaces.nsIAbCard);
      var nCard = tCards.currentItem().QueryInterface(Components.interfaces.nsIAbCard);
if (tCard.isAnEmailAddress || tCard.isASpecialGroup || !tCard.isANormalCard)
alert('unnormal card: '+tCard.displayName+'('+tCard.primaryEmail+') isAnEmailAddress='+tCard.isAnEmailAddress
+' isANormalCard='+tCard.isANormalCard+' isASpecialGroup='+tCard.isASpecialGroup);
            //tCard.isANormalCard is also true for maillist!
      if (!tCard.isMailList) { // only non-maillists
//        var nCard=cur.addCard(tCard);  // add card, afterwards: nCard.equals(tCard) is true !!
        cur.dropCard(nCard, false);  // changes nCard to be the new card!
        //find maillists, the card is in
        var tMailLists=temp.childNodes; //mailLists=nsISimpleEnumerator, Get the children directories
        while (tMailLists.hasMoreElements()) {
          var tMailList=tMailLists.getNext().QueryInterface(Components.interfaces.nsIAbDirectory);
          // loop over cards in maillist
          for (var i=0; i<tMailList.addressLists.Count(); i++)
          {
            var tmlCard = tMailList.addressLists.GetElementAt(i).QueryInterface(Components.interfaces.nsIAbCard);
            if (tCard.equals(tmlCard)) {
              // find new maillist and insert new card
              var nMailLists=cur.childNodes; //mailLists=nsISimpleEnumerator, Get the children directories
              while (nMailLists.hasMoreElements()) {
                var nMailList=nMailLists.getNext().QueryInterface(Components.interfaces.nsIAbDirectory);
                if( tMailList.dirName == nMailList.dirName) {  // got it!
                  nMailList.dropCard(nCard, false);  // insert new card to new maillist
                  break;
                }
              }
              break;
            }
          }
        }
      }
      try { tCards.next() } catch(e) { goon=0; }
    }
  } catch(e) { alert(e); }

  tdb.forceClosed();  //!!!

  return '';
}

function getLastModified(dir) {  //dir=nsIAbDirectory
  if (!dir) return 0;

  var lastModifiedDate=dir.lastModifiedDate;  //set via abs_gAbSessionListener
dump('abs: last modified date of dir '+dir.dirName+': '+lastModifiedDate+'\n');
  var count=0;
  var email='';
  try  {
    var cardsE=dir.childCards;
    var cards = Components.classes["@mozilla.org/supports-array;1"].
      getService(Components.interfaces.nsISupportsArray);
      //convert nsIEnumerator to nsISupportsArray (don't try dir.dropCard()!)
    var goon=1;
      // Loop over the cards
    try { cardsE.first() } catch (e) { goon=0; }
    while (goon) {
      var card = cardsE.currentItem().QueryInterface(Components.interfaces.nsIAbCard);
      var lmt=card.lastModifiedDate;
      if (lmt>lastModifiedDate) {
        lastModifiedDate=lmt;
        email=card.primaryEmail;
      }
      count++;
      try { cardsE.next() } catch(e) { goon=0; }
    }
  } catch(e) { alert(e); }
if (email) dump('abs: last modified date of email '+email+': '+lastModifiedDate+'\n');
  return lastModifiedDate;
}

/*
 * return URI for remote syncronization
 */
function getSyncURI(masked)
{
  var syncprotocol;
  var synchost='';
  var syncuser='';
  var syncpassword='';
  var syncpath='';
  try{ syncprotocol = pref.getCharPref( "extensions.addressbookssync.protocol" ); } catch(e){}
  try{ synchost = pref.getCharPref( "extensions.addressbookssync.host" ); } catch(e){}
  try{ syncuser = pref.getCharPref( "extensions.addressbookssync.user" ); } catch(e){}
  try{ syncpassword = get_password(synchost, syncuser); } catch(e){}
  try{ syncpath = pref.getCharPref( "extensions.addressbookssync.path" ); } catch(e){}

  var pwd='';
  if (masked) {
    var pwl=syncpassword.length;
    for(var i=0; i<pwl;i++) pwd += "*";
  } else
    pwd=syncpassword;

  if( ! synchost )
    return "";
  var userpass=syncuser +(pwd?":":"")+pwd;
  if(userpass)
    userpass += "@";
  return syncprotocol + "://" + userpass + synchost + syncpath+'/';
}

/*
 *  get a special Directory, 'ProfD' or 'TmpD'
 */
function getDir(special) {
/*
  var profileFolder = directoryService.get(special,  Components.interfaces.nsIFile);
  var profileDir = profileFolder.QueryInterface(Components.interfaces.nsILocalFile);
  var nsIMabFile = profileDir.clone().QueryInterface(Components.interfaces.nsILocalFile);
*/
  var dir = gDirectoryService.get(special,  Components.interfaces.nsIFile)
    .QueryInterface(Components.interfaces.nsILocalFile);
//  return dir.clone().QueryInterface(Components.interfaces.nsILocalFile);
  return dir;
}

function deleteTempFile(filename) {
  var nsITmpFile=getDir('TmpD');
  nsITmpFile.appendRelativePath(filename);
  if (!nsITmpFile.exists() || !nsITmpFile.isFile()) return true;
  var deleted=false;
  try { nsITmpFile.remove(false) }
  catch(e) {
    switch (e.result) {
      case Components.results.NS_ERROR_FILE_ACCESS_DENIED:
    //  after call to replaceAddressBook, an immediate remove
    //  throws an 'access denied'
//        setTimeout(deleteTempFile, 1000, filename, count);
        break;
      case Components.results.NS_ERROR_FILE_NOT_FOUND:
        deleted=true;
        break;
      default:
        alert(e.message);
        deleted=true; // not really
        break;
    }
    return deleted;
  }
  return true;
}

function getTimeStamp()
{
    var date = new Date();
    return Math.floor( date.getTime() / 1000 );

} // getTimeStamp


function errorstr(aError)
{
  var error;

  var intErr=0;
  if (typeof aError == 'string') {
    var err=Number(aError);
    if (!isNaN(err)) aError=err;
  }
  //Errorcodes:
  //    xpcom/base/nsError.h
  //    netwerk/base/public/nsNetError.h
  switch(aError){
  // internal errors
    case -1:
      intErr=1;
      error=strings.getString('unexperr'); // 'unexpected initialization error';
      break;
    case -2:
      intErr=2;
      error=strings.getString('localreaderr'); // 'local mab file read error';  //access denied (also gives unhandled exception), does not exist
      break;
    case -3:
      intErr=3;
      error=strings.getString('illimap');
      break;
  //Thunderbird network error, see netwerk/base/public/nsNetError.h
    case 0x804b000d:
      //NS_ERROR_CONNECTION_REFUSED 13
      //The connection attempt failed, for example, because no server was listening
      //at specified host:port.
      intErr=4;
      error=strings.getString('protnotavail'); // 'protocol not available';
      break;
    case 0x804b000e:
        //NS_ERROR_NET_TIMEOUT 14
        //The connection was lost due to a timeout error.
      intErr=12;
      error=strings.getString('timeout'); // 'timeout';
      break;
    case 0x804b0011:
      //NS_ERROR_NO_CONTENT 17
      intErr=16;
      error='No content';
        // used, when Firefox is set as 'network.protocol-handler.app.ftp'
        // for FTP. The addressbook is shown in Firefox!
      break;
    case 0x804b0012:
      //NS_ERROR_UNKNOWN_PROTOCOL 18
      // The URI scheme corresponds to an unknown protocol handler.
      intErr=17;
      error='Unknown protocol handler';
      // happens after removing application-handler after NS_ERROR_NO_CONTENT
      break;
    case 0x804b0015:
      //NS_ERROR_FTP_LOGIN 21
      intErr=10;
      error=strings.getString('illcred'); // illegal credentials';
      break;
    case 0x804b0016:
      //NS_ERROR_FTP_CWD 22
      intErr=15;
      error=strings.getString('nofile');
      break;
//NS_ERROR_FTP_PASV 23
//NS_ERROR_FTP_PWD 24
//NS_ERROR_FTP_LIST 25
    case 0x804b001e:
      //NS_ERROR_UNKNOWN_HOST 30
      intErr=5;
      error=strings.getString('hostnotfound'); // 'host not found';
      break;
    case 0x804b002a:
      //NS_ERROR_UNKNOWN_PROXY_HOST 42
      intErr=13;
      error=strings.getString('proxynotfound'); // 'proxy not found';
      break;
    case 0x804b0048:
      //NS_ERROR_PROXY_CONNECTION_REFUSED 72
      intErr=14;
      error=strings.getString('proxyerr'); // 'proxy error';
      break;
  //external errors (server error, windows error)
    case 0x80004002:
      /* NS_ERROR_NO_INTERFACE: Returned when a given interface is not supported. */
      intErr=18;
      error='External application handler set';
        //This happens on ftp upload when an external handler is set
      break;
    case 0x80004005:
      /* NS_ERROR_FAILURE: Returned when a function fails */
      intErr=6;
      error=strings.getString('permdenied'); // frher 'fileunavail', file unavailable
        //??? (with FTP: Error 550: File unavailable (e.g., file not found, no access).
        //(with IMAP: wenn der alte Upload mit 'CopyFileMessage' verwendet wird)
      break;
    case 405:
      intErr=7;
      error=strings.getString('methodnotallowed'); // method not allowed';  //method is something like PUT,POST or GET
      break;
    case 404:
      intErr=8;
      error=strings.getString('notfound'); // not found';
      break;
    case 403:
      intErr=9;
      error=strings.getString('permdenied'); // permission denied';
      break;
    case 401:
      intErr=11;
      error=strings.getString('illcred'); // illegal credentials';
      break;
    case 0:
    case 201:
    case 204:
      error=strings.getString('ok');
      break;
    default:
      var sError;
      if (aError&0xffff0000) {
        if ((aError>>>16)==0x804b)  //TB Netzwerk errors, see netwerk/base/public/nsNetError.h
          sError='TB-'+(aError&0xffff).toString(10);
        else
          sError='0x'+aError.toString(16);
      } else
        sError=aError;
      error='Error '+sError;
      break;
  }
//  throws 'no memory' exception
//  var errorService = Components.classes["@mozilla.org/xpcom/error-service;1"].
//            getService(Components.interfaces.nsIErrorService);
//  error+=' ('+errorService.getErrorStringBundleKey(aError)+')';
  if (intErr) error+=' ('+intErr+')';
  return error;
}

var allStatus;
function statustxt(aStatus, newline, doalert)
{
  var usespopup=window.location=='chrome://addressbookssync/content/download.xul'
    || window.location=='chrome://addressbookssync/content/upload.xul';
  if (gUsesPopup) {
    if (newline) document.getElementById("status").value += "\n";
    var status=document.getElementById("status");
    status.value += aStatus;
    status.inputField.scrollTop=status.inputField.scrollHeight;
    if (doalert) gLifeTime=5; //show window a little bit longer
  } else {
if (newline) allStatus += "\n"; //used for debugging
allStatus+=aStatus;
    var maindoc = window.location!='chrome://messenger/content/messenger.xul'?window.opener.document:document;
    if (newline) maindoc.getElementById('statusText').label=aStatus
    else maindoc.getElementById('statusText').label+=aStatus;
    if (doalert)
      alert(maindoc.getElementById('statusText').label);
  }
}

function addMenu()
{
  try{
    var menu=document.getElementById("tasksMenu");
    var popup=menu.firstChild;
    var sep=document.getElementById("prefSep");
    var strings = document.getElementById("strings");
    var label=strings.getString('menuname'); //"Addressbooks Synchronizer";
    var accesskey=strings.getString('menushortcut');
  }catch(e){alert(e); return;}

  var  element = document.createElementNS("http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul", "menuitem");
  element.setAttribute("id", "gg-addressbookSync");
  element.setAttribute("label", label);
  element.setAttribute("accesskey", accesskey);
  element.setAttribute("oncommand", "showOptions();");
  popup.insertBefore(element,sep);
}

//There is a problem, when deleting and recreating existing addressbooks
//(syncpolicy=='file'):
//If the addressbook window is open, tb may crash afterwards if accessing
//the recreated addressbook in the addressbook window.
//therefore we close the addressbook or select the first addressbook in the window
//and delay the delete/create process
//But nothing really helped :-(
function secureAddressbook() {

  var count=0;
  var windowsEnum = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService()
                   .QueryInterface(Components.interfaces.nsIWindowMediator)
                   .getEnumerator("mail:addressbook");
  while (windowsEnum.hasMoreElements()) {
    var w=windowsEnum.getNext();
//    w.close();
    var abTree=w.document.getElementById("dirTree");
    if (abTree) {
      if (abTree.view.selection.currentIndex!=0) {
        w.SelectFirstAddressBook();
        if (w.GetSelectedDirectory()!=kPersonalAddressbookURI) {
          alert('Not personal addressbook - please inform the Author (G.Gersdorf@ggbs.de)');
        }
        count++;
      }
    } else {
//alert('no abTree :-((');
    }
  }
  return count;
}

function showOptions()
{
  var windowsEnum = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService()
                   .QueryInterface(Components.interfaces.nsIWindowMediator)
                   .getEnumerator("addressbookssync-options");
  var options=windowsEnum.getNext();
  if (options)
    options.focus();
  else
    window.open("chrome://addressbookssync/content/options.xul","_blank","chrome,resizable,titlebar=yes");
}

/*
 *  This is, where every up-/download starts!
 */
function showPopup(direction, singlebook, usepopup, automatic, force)
{
  //singlebook: a mabFile or null
  //usepopup: if true, use popups, else ask property 'hidepopups'
  //automatic: true, if its an automatic down-/upload and programmstart/-end
  //force: if true, upload mabs, even if not changed

  if (!singlebook && syncedAddressbooks((direction=='download')?'down':'up').length==0)
    return;  //nothing to do

//TODO: possible race condition!
  if (gAddressbooksSync.state!='') {  // we are already doing an up-/download
    if (!automatic) {
      var strings = document.getElementById("strings");
      alert(strings.getString('inprogress'));
    }
    return;
  }

  // see comment with secureAddressbook()
  if (direction=='download' && secureAddressbook()) {
    setTimeout(showPopup, 1000, direction, singlebook, usepopup, automatic, force);
    return;
  }

  gAddressbooksSync.state=direction;
dump('direction is '+direction+', state now '+gAddressbooksSync.state+'\n');

  var hidepopups=false;
  if (!usepopup)
    try{ hidepopups = pref.getBoolPref( "extensions.addressbookssync.hidepopups" ); }
      catch(e){ alert(e); hidepopups=false; }
  if (hidepopups) {
    if (direction=='download') download(singlebook, force);    //in download.js
    else                       upload(singlebook, force);      //in upload.js
  } else {
    var x=Math.floor(window.outerWidth/2+window.screenX);
    var y=Math.floor(window.outerHeight/2+window.screenY);
    var w=window.openDialog("chrome://addressbookssync/content/"+direction+".xul","_blank",
      'chrome,resizable,titlebar=no', singlebook, force);
//if adding more arguments: update in call to upload() and download() above
//    AND in upload.xul and download.xul!!

    w.screenX=x;
    w.screenY=y;
  }
}

function checkOffline(direction) {
  if (ioService.offline) {
    var windowsEnum = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService()
                     .QueryInterface(Components.interfaces.nsIWindowMediator)
                     .getEnumerator("mail:3pane");  //.getMostRecentWindow("mail:3pane")
    var w=windowsEnum.getNext();  // there must be one (the current)! (really? what if tb killed)
    if (!w) {
      gAbort=2;
      killMe();
      return false;
    }
    var goOnline = direction=='upload' ?
        w.PromptSendMessagesOffline() : w.PromptGetMessagesOffline();
    if(goOnline==0) {
      if (!w.gOfflineManager)
        w.GetOfflineMgrService();
      w.gOfflineManager.goOnline(false /* sendUnsentMessages */,
                             false /* playbackOfflineImapOperations */,
                             w.msgWindow);
    } else {
      gAbort=2;
      killMe();
      return false;
    }
  }
  return true;
}

function killMe()
{
  if (gAbort<2 && cleanUp) {
    cleanUp();
    cleanUp=null;
  }
  if (gTempName && !gAbort) {
    if (!deleteTempFile(gTempName)) {
      setTimeout(killMe,1000);
      return;
    } else
    gTempName='';
  }
  if (gAbort==1)
    // this does not fire on unexpected errors, since
    // killMe is probably not called from processing loop
    statustxt(strings.getString('aborted'), true, false);

  if(gLifeTime>0 && gUsesPopup && gAbort<2) {
    gLifeTime--;
    setTimeout(killMe,1000);
  } else {
    var observerService = Components.classes["@mozilla.org/observer-service;1"]
      .getService(Components.interfaces.nsIObserverService);
    gAddressbooksSync.state=''; //Up-/Download ready
    statustxt('', true, false);
    if (gUsesPopup) setTimeout(close,0);
  }
}


function addListener() {
  var addrbookSession = Components.classes["@mozilla.org/addressbook/services/session;1"]
      .getService().QueryInterface(Components.interfaces.nsIAddrBookSession);
  addrbookSession.addAddressBookListener(
    abs_gAbSessionListener, -1);
//    Components.interfaces.nsIAddrBookSession.directoryRemoved |
//   Components.interfaces.nsIAddrBookSession.changed);
}
function removeListener() {
  var addrbookSession = Components.classes["@mozilla.org/addressbook/services/session;1"]
      .getService().QueryInterface(Components.interfaces.nsIAddrBookSession);
  addrbookSession.removeAddressBookListener(abs_gAbSessionListener);
}
var abs_gAbSessionListener = {
  onItemAdded: function(parentDir, item) {
    var dir=parentDir.QueryInterface(Components.interfaces.nsIAbDirectory);
    try {
      var card=item.QueryInterface(Components.interfaces.nsIAbCard);
dump('abs: item added parent='+dir.dirName+' email='+card.primaryEmail+'\n');
      dir.lastModifiedDate=Math.round(Date.now()/1000);
      card.setCardValue('LastModifiedDate',dir.lastModifiedDate);
    } catch(e) {}
  },
  onItemRemoved: function(parentDir, item) {
    var dir=parentDir.QueryInterface(Components.interfaces.nsIAbDirectory);
dump('abs: something removed parent='+dir.dirName+'\n');
    try {
      var card=item.QueryInterface(Components.interfaces.nsIAbCard);
dump('abs: item removed parent='+dir.dirName+' email='+card.primaryEmail+'\n');
    } catch(e) {}
    dir.lastModifiedDate=Math.round(Date.now()/1000);
  },
  onItemPropertyChanged: function(item, property, oldValue, newValue) {
    //property, oldValue and newValue are always null
    var card=item.QueryInterface(Components.interfaces.nsIAbCard);
dump('item changed '+card.primaryEmail+' '+property+': '+oldValue+'->'+newValue+'\n');
    // this also sets the lastModifiedDate on the card
  }
};
